#include <iostream>
#include <cstdlib>
#include <sys/ioctl.h>

using namespace std;

namespace console
{
  int width = 80;
  int height = 26;

  enum colors { BLACK = 30, GREEN = 32, WHITE = 37 };

  void set_pos(int col, int row)
  {
    cout << "\033[" << (row+1) << ";" << (col+1) << "f";
  }

  void set_color(bool bright, int color)
  {
    cout << "\033[" << bright << ";" <<color << "m";
  }
  
  unsigned char *screen;

  void clear()
  {
    for(int i = 0; i < width; i++) for(int j = 0; j < height; j++) 
    {
      screen[i + j * width] = 255;
    }
  }
 
  void init(int iwidth, int iheight)
  {
    width = iwidth;
    height = iheight;
    screen = new unsigned char[width * height];
    clear();
  }
}

namespace matrix
{
  class thread
  {
    public:
      thread()
      {
        init();
      }
      ~thread()
      {
      }
      void init()
      {
        pos_row = -1;
        pos_col = rand() % console::width;
        speed  = 1  + (rand()%9);
        length = 10 + console::height / speed;
        wait   = rand() % 100;
        time_trnsf = 5 + (rand() % 5);
        time = 0;
        refresh_p = rand() % console::height;
        refresh_speed = 1 + (rand()%3);
        content = new char[console::height];
        refresh();
      }
      void step(void)
      {
        if ((time % speed) == 0)
        {
          if (time > wait)
          {
            pos_row++;
            if ((time % time_trnsf) == 0) refresh();
          }
          if (pos_row - length > console::height) init();
        }
        refresh_p--;
        if (refresh_p < 0) refresh_p =  console::height - 1;
        content[refresh_p] = newchar(12);

        time++;
      }
      void draw(void)
      {
        int pos = pos_row;
        for (int i = 0; i < length; i++) 
        {
          console::set_color((i==0), 32);
          pos = pos_row - i;
          if ((pos >= 0) 
              && (pos < console::height)
              && (console::screen[pos_col + pos * console::width] > pos_row))
          {
            console::set_pos(pos_col, pos);
            console::screen[pos_col + pos * console::width] = pos_row;
            cout << content[pos]; 
          }          
        }
        if ((pos-1 >= 0) && (pos-1 < console::height))
        {
          console::set_pos(pos_col, pos-1);
          cout << " "; 
          console::screen[pos_col + (pos-1) * console::width] = 255;
        }
      }
    private:
      unsigned char newchar(int posib)
      {
        return(   (rand() % posib == 0) 
                ? (' ')
                : ('0' + (rand()%10))
              );
      }
      void refresh(void)
      {
        for (int i = 0; i < console::height; i++) 
        {
          content[i] = newchar(12);
        }
      }
      int time_trnsf;
      int pos_row;
      int pos_col;
      int length;
      int speed;
      int time;  
      int wait;
      int refresh_p;
      int refresh_speed;
      char *content;
  };
}

int main()
{
  bool ret;
  int  lines, columns;
  struct winsize win;

  if(ioctl (STDOUT_FILENO, TIOCGWINSZ, &win) == 0)
  {
    lines   = win.ws_row;
    columns = win.ws_col;
  }
  else
  {
    fprintf(stderr,"Can not determine terminal size\n");
    lines   = 25;
    columns = 80;
  }

  console::init(columns, lines);
  int thread_num = console::width * 2 ;
  
  matrix::thread *m = new matrix::thread[thread_num];

  while(1)
  {  
    console::clear();
    for (int i = 0; i < thread_num; i++)
    {  
      m[i].step();
      m[i].draw();
    }

    usleep(50000);
  }
}
